@@ -1,32 +1,18 @@
-/* $Id: cache.c,v 1.3.4.1 2006/01/11 01:59:17 canacar Exp $ */
+/* $Id: cache.c,v 1.5 2007/06/09 23:43:27 canacar Exp $ */
 /*
- * Copyright (c) 2001 Can Erkin Acar
- * All rights reserved.
+ * Copyright (c) 2001, 2007 Can Erkin Acar <canacar@openbsd.org>
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- *
- *    - Redistributions of source code must retain the above copyright
- *      notice, this list of conditions and the following disclaimer.
- *    - Redistributions in binary form must reproduce the above
- *      copyright notice, this list of conditions and the following
- *      disclaimer in the documentation and/or other materials provided
- *      with the distribution.
- *
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
- * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
- * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
- * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
- * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
- * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
- * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
- * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
- * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
- * POSSIBILITY OF SUCH DAMAGE.
+ * Permission to use, copy, modify, and distribute this software for any
+ * purpose with or without fee is hereby granted, provided that the above
+ * copyright notice and this permission notice appear in all copies.
  *
+ * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
+ * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
+ * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
+ * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
+ * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
+ * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
 #include <sys/types.h>
@@ -48,8 +34,8 @@
 #include "cache.h"
 
 /* prototypes */
-void update_state(struct sc_ent *, struct pf_state *, double);
-struct sc_ent *cache_state(struct pf_state *);
+void update_state(struct sc_ent *, pf_state_t *, double);
+struct sc_ent *cache_state(pf_state_t *);
 static __inline int sc_cmp(struct sc_ent *s1, struct sc_ent *s2);
 
 /* initialize the tree and queue */
@@ -101,22 +87,22 @@ cache_init(int max)
 }
 
 void
-update_state(struct sc_ent *prev, struct pf_state *new, double rate)
+update_state(struct sc_ent *prev, pf_state_t *new, double rate)
 {
 	assert (prev != NULL && new != NULL);
 	prev->t = time(NULL);
 	prev->rate = rate;
 #ifdef HAVE_INOUT_COUNT
-	prev->bytes = new->bytes[0] + new->bytes[1];
+	prev->bytes = COUNTER(new->bytes[0]) + COUNTER(new->bytes[1]);
 #else
-	prev->bytes = new->bytes;
+	prev->bytes = COUNTER(new->bytes);
 #endif
 	if (prev->peak < rate)
 		prev->peak = rate;
 }
 
 void
-add_state(struct pf_state *st)
+add_state(pf_state_t *st)
 {
 	struct sc_ent *ent;
 	assert(st != NULL);
@@ -139,7 +125,7 @@ add_state(struct pf_state *st)
 	ent->af = st->af;
 	ent->proto = st->proto;
 #ifdef HAVE_INOUT_COUNT
-	ent->bytes = st->bytes[0] + st->bytes[1];
+	ent->bytes = COUNTER(st->bytes[0]) + COUNTER(st->bytes[1]);
 #else
 	ent->bytes = st->bytes;
 #endif
@@ -153,7 +139,7 @@ add_state(struct pf_state *st)
 
 /* must be called only once for each state before cache_endupdate */
 struct sc_ent *
-cache_state(struct pf_state *st)
+cache_state(pf_state_t *st)
 {
 	struct sc_ent ent, *old;
 	double sd, td, r;
@@ -176,10 +162,10 @@ cache_state(struct pf_state *st)
 	}
 
 #ifdef HAVE_INOUT_COUNT
-	if (st->bytes[0] + st->bytes[1] < old->bytes)
+	if (COUNTER(st->bytes[0]) + COUNTER(st->bytes[1]) < old->bytes)
 		return (NULL);
 
-	sd = st->bytes[0] + st->bytes[1] - old->bytes;
+	sd = COUNTER(st->bytes[0]) + COUNTER(st->bytes[1]) - old->bytes;
 #else
 	if (st->bytes < old->bytes)
 		return (NULL);
