@@ -1,4 +1,4 @@
-/* $Id: engine.c,v 1.4 2007/06/09 23:43:27 canacar Exp $	 */
+/* $Id: engine.c,v 1.13 2007/11/07 06:31:32 canacar Exp $	 */
 /*
  * Copyright (c) 2001, 2007 Can Erkin Acar <canacar@openbsd.org>
  *
@@ -27,8 +27,6 @@
 #include <string.h>
 #include <unistd.h>
 
-#include "config.h"
-
 #include "engine.h"
 
 #ifndef MIN
@@ -80,6 +78,7 @@ char tmp_buf[MAX_LINE_BUF];
 char cmdbuf[MAX_LINE_BUF];
 int cmd_len = -1;
 struct command *curr_cmd = NULL;
+char *curr_message = NULL;
 
 void print_cmdline(void);
 
@@ -94,6 +93,7 @@ tb_start(void)
 {
 	tb_ptr = tmp_buf;
 	tb_len = sizeof(tmp_buf);
+	tb_ptr[0] = '\0';
 }
 
 void
@@ -245,14 +245,118 @@ print_fld_str(field_def *fld, const char *str)
 	}
 }
 
+void
+print_bar_title(field_def *fld)
+{
+	char buf[16];
+	int len, div, i, tr, tw, val, pos, cur;
+
+	int divs[] = {20, 10, 5, 4, 3, 2, 1, 0};
+
+	if (fld->width < 1)
+		return;
+
+	len = snprintf(buf, sizeof(buf), " %d\\", fld->arg);
+	if (len >= sizeof(buf))
+		return;
+
+	for (i = 0; divs[i]; i++)
+		if (divs[i] * len <= fld->width)
+			break;
+
+	if (divs[i] == 0) {
+		print_fld_str(fld, "*****");
+		return;
+	}
+
+	div = divs[i];
+
+	val = 0;
+	pos = 0;
+	tr = fld->arg % div;
+	tw = fld->width % div;
+
+	tb_start();
+	cur = 0;
+	for(i = 0; i < div; i++) {
+		tw += fld->width;
+		tr += fld->arg;
+
+		while (tr >= div) {
+			val++;
+			tr -= div;
+		}
+		while (tw >= div) {
+			pos++;
+			tw -= div;
+		}
+
+		len = snprintf(buf, sizeof(buf), "%d\\", val);
+		while (cur < pos - len) {
+			tbprintf(" ");
+			cur++;
+		}
+		tbprintf("%s", buf);
+		cur += len;
+	}
+
+	print_fld_tb(fld);
+}
+
+void
+print_fld_bar(field_def *fld, int value)
+{
+	int i, tw, val, cur;
+
+	if (fld->width < 1)
+		return;
+
+	val = 0;
+	tw = fld->arg / 2;
+
+	tb_start();
+	cur = 0;
+	for(i = 0; i < fld->width; i++) {
+		tw += fld->arg;
+
+		while (tw >= fld->width) {
+			val++;
+			tw -= fld->width;
+		}
+		if (val > value)
+			break;
+		tbprintf("#");
+	}
+
+	print_fld_tb(fld);
+}
+
+void
+print_fld_tb(field_def *fld)
+{
+	print_fld_str(fld, tmp_buf);
+	tb_end();
+}
+
 void
 print_title(void)
 {
 	field_def **fp;
 
 	if (curr_view != NULL && curr_view->view != NULL) {
-		for (fp = curr_view->view; *fp != NULL; fp++)
-			print_fld_str(*fp, (*fp)->title);
+		for (fp = curr_view->view; *fp != NULL; fp++) {
+			switch((*fp)->align) {
+			case FLD_ALIGN_LEFT:
+			case FLD_ALIGN_RIGHT:
+			case FLD_ALIGN_CENTER:
+			case FLD_ALIGN_COLUMN:
+				print_fld_str(*fp, (*fp)->title);
+				break;
+			case FLD_ALIGN_BAR:
+				print_bar_title(*fp);
+				break;
+			}
+		}
 	}
 	end_line();
 }
@@ -412,22 +516,41 @@ add_view(field_view *fv)
 		set_curr_view(ent);
 }
 
-void
+int
 set_view(char *opt)
 {
-	struct view_ent *ve;
+	struct view_ent *ve, *vm = NULL;
 	field_view *v;
+	int len;
 
-	if (opt == NULL)
-		return;
+	if (opt == NULL || (len = strlen(opt)) == 0)
+		return 1;
 
 	CIRCLEQ_FOREACH(ve, &view_head, entries) {
 		v = ve->view;
-		if (strcasecmp(opt, v->name) == 0) {
-			set_curr_view(ve);
-			return;
+		if (strncasecmp(opt, v->name, len) == 0) {
+			if (vm)
+				return 1;
+			vm = ve;
 		}
 	}
+
+	if (vm) {
+		set_curr_view(vm);
+		return 0;
+	}
+
+	return 1;
+}
+
+void
+foreach_view(void (*callback)(field_view *))
+{
+	struct view_ent *ve;
+
+	CIRCLEQ_FOREACH(ve, &view_head, entries) {
+		callback(ve->view);
+	}
 }
 
 int
@@ -531,8 +654,7 @@ print_fld_age(field_def *fld, unsigned int age)
 	return;
 	
  ok:
-	print_fld_str(fld, tmp_buf);
-	tb_end();
+	print_fld_tb(fld);
 }
 
 void
@@ -576,8 +698,7 @@ err:
 	return;
 
 ok:
-	print_fld_str(fld, tmp_buf);
-	tb_end();
+	print_fld_tb(fld);
 }
 
 void
@@ -622,7 +743,8 @@ print_fld_uint(field_def *fld, unsigned int size)
 	if (tbprintf("%u", size) > len)
 		print_fld_str(fld, "*");
 	else
-		print_fld_str(fld, tmp_buf);
+		print_fld_tb(fld);
+	tb_end();
 }
 
 
@@ -744,8 +866,13 @@ disp_update(void)
 			lines = curr_mgr->header_fn();
 			if (lines < 0)
 				return (1);
+#ifdef HOME_BOTTOM
+			curr_line = ++lines;
+			home_line = lines + maxprint + 1;
+#else
 			home_line = lines++;
 			curr_line = lines;
+#endif
 		}
 
 		print_title();
@@ -825,29 +952,71 @@ setup_term(int dmax)
 }
 
 struct command *
-command_set(struct command *cmd)
+command_set(struct command *cmd, const char *init)
 {
 	struct command *prev = curr_cmd;
 
 	if (cmd) {
-		cmd_len = 0;
-		cmdbuf[0] = 0;
+		if (init) {
+			cmd_len = strlcpy(cmdbuf, init, sizeof(cmdbuf));
+			if (cmd_len >= sizeof(cmdbuf)) {
+				cmdbuf[0] = '\0';
+				cmd_len = 0;
+			}
+		} else {
+			cmd_len = 0;
+			cmdbuf[0] = 0;
+		}
 	}
+	curr_message = NULL;
 	curr_cmd = cmd;
 	need_update = 1;
 	return prev;
 }
 
+const char *
+message_set(const char *msg) {
+	char *prev = curr_message;
+	if (msg)
+		curr_message = strdup(msg);
+	else
+		curr_message = NULL;
+	free(prev);
+	return NULL;
+}
+
+int
+msgprintf(char *format, ...)
+	GCC_PRINTFLIKE(1,2)       /* defined in curses.h */
+{
+	static char buf[1024];
+	int len;
+	va_list arg;
+
+	va_start(arg, format);
+	len = vsnprintf(buf, sizeof(buf), format, arg);
+	va_end(arg);
+	buf[sizeof(buf) - 1] = '\0';
+
+	if (rawmode)
+		fprintf(stderr, "%s", buf);
+	else
+		message_set(buf);
+	return len;
+}
+
 void
 print_cmdline(void)
 {
-	if (curr_cmd == NULL)
-		return;
-
-	attron(A_STANDOUT);
-	mvprintw(home_line, 0, "%s: ", curr_cmd->prompt);
-	attroff(A_STANDOUT);
-	printw("%s", cmdbuf);
+	if (curr_cmd) {
+		attron(A_STANDOUT);
+		mvprintw(home_line, 0, "%s: ", curr_cmd->prompt);
+		attroff(A_STANDOUT);
+		printw("%s", cmdbuf);
+	} else if (curr_message) {
+		mvprintw(home_line, 0, "> %s", curr_message);
+	}
+	clrtoeol();
 }
 
 
@@ -861,7 +1030,6 @@ cmd_keyboard(int ch)
 		if (cmd_len < sizeof(cmdbuf) - 1) {
 			cmdbuf[cmd_len++] = ch;
 			cmdbuf[cmd_len] = 0;
-			need_update = 1;
 		} else
 			beep();
 	}
@@ -871,7 +1039,7 @@ cmd_keyboard(int ch)
 	case 0x0a:
 	case 0x0d:
 	{
-		struct command * c = command_set(NULL);
+		struct command * c = command_set(NULL, NULL);
 		c->exec();
 		break;
 	}
@@ -880,10 +1048,17 @@ cmd_keyboard(int ch)
 	case CTRL_H:
 		if (cmd_len > 0) {
 			cmdbuf[--cmd_len] = 0;
-			need_update = 1;
 		} else
 			beep();
 		break;
+	case 0x1b:
+	case CTRL_G:
+		if (cmd_len > 0) {
+			cmdbuf[0] = '\0';
+			cmd_len = 0;
+		} else
+			command_set(NULL, NULL);
+		break;
 	default:
 		break;
 	}
@@ -898,6 +1073,7 @@ keyboard(void)
 
 	if (curr_cmd) {
 		cmd_keyboard(ch);
+		print_cmdline();
 		return;
 	}
 
@@ -906,6 +1082,13 @@ keyboard(void)
 			if (curr_mgr->key_fn(ch))
 				return;
 
+	if (curr_message != NULL) {
+		if (ch > 0) {
+			curr_message = NULL;
+			need_update = 1;
+		}
+	}
+
 	switch (ch) {
 	case ' ':
 		gotsig_alarm = 1;
