@@ -1,4 +1,4 @@
-/* $Id: pftop.c,v 1.80 2007/06/09 23:43:27 canacar Exp $	 */
+/* $Id: pftop.c,v 1.91 2007/11/07 06:33:14 canacar Exp $	 */
 /*
  * Copyright (c) 2001, 2007 Can Erkin Acar
  * Copyright (c) 2001 Daniel Hartmeier
@@ -65,6 +65,7 @@
 
 #include "engine.h"
 #include "cache.h"
+#include "sf-gencode.h"
 
 
 #define MIN_NUM_STATES 1024
@@ -163,10 +164,15 @@ pf_state_t *state_buf = NULL;
 int state_buf_len = 0;
 u_int32_t *state_ord = NULL;
 u_int32_t num_states = 0;
+u_int32_t num_states_all = 0;
 u_int32_t num_rules = 0;
 u_int32_t num_queues = 0;
 int cachestates = 0;
 
+struct bpf_program filter_prog;
+char *filter_string = NULL;
+int dumpfilter = 0;
+
 #ifndef HAVE_RULE_LABELS
 #define PF_RULE_LABEL_SIZE 20
 #endif
@@ -176,42 +182,42 @@ int cachestates = 0;
 
 /* Define fields */
 field_def fields[] = {
-	{"SRC", 20, 45, 1, FLD_ALIGN_LEFT, -1, 0, 0},
-	{"DEST", 20, 45, 1, FLD_ALIGN_LEFT, -1, 0, 0},
-	{"GW", 20, 45, 1, FLD_ALIGN_LEFT, -1, 0, 0},
-	{"STATE", 5, 23, 18, FLD_ALIGN_COLUMN, -1, 0, 0},
-	{"AGE", 5, 9, 4, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"EXP", 5, 9, 4, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"PR ", 4, 9, 1, FLD_ALIGN_LEFT, -1, 0, 0},
-	{"DIR", 1, 3, 2, FLD_ALIGN_CENTER, -1, 0, 0},
-	{"PKTS", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"BYTES", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"RULE", 2, 4, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"LABEL", MIN_LABEL_SIZE, MIN_LABEL_SIZE, 1, FLD_ALIGN_LEFT, -1, 0, 0},
-	{"STATES", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"EVAL", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"ACTION", 1, 8, 4, FLD_ALIGN_LEFT, -1, 0, 0},
-	{"LOG", 1, 3, 2, FLD_ALIGN_LEFT, -1, 0, 0},
-	{"QUICK", 1, 1, 1, FLD_ALIGN_LEFT, -1, 0, 0},
-	{"KS", 1, 1, 1, FLD_ALIGN_LEFT, -1, 0, 0},
-	{"IF", 4, 6, 1, FLD_ALIGN_LEFT, -1, 0, 0},
-	{"INFO", 40, 80, 1, FLD_ALIGN_LEFT, -1, 0, 0},
-	{"MAX", 3, 5, 2, FLD_ALIGN_RIGHT, -1, 0, FLD_FLAG_HIDDEN},
-	{"RATE", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"AVG", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"PEAK", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"ANCHOR", 6, 16, 1, FLD_ALIGN_LEFT, -1, 0, FLD_FLAG_HIDDEN},
-	{"QUEUE", 15, 30, 1, FLD_ALIGN_LEFT, -1, 0, 0},
-	{"BW", 4, 5, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"SCH", 3, 4, 1, FLD_ALIGN_LEFT, -1, 0, 0},
-	{"PRIO", 1, 4, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"DROP_P", 6, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"DROP_B", 6, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"QLEN", 4, 4, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"BORROW", 4, 6, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"SUSPENDS", 4, 6, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"P/S", 3, 7, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"B/S", 4, 7, 1, FLD_ALIGN_RIGHT, -1, 0, 0}
+	{"SRC", 20, 45, 1, FLD_ALIGN_LEFT, -1, 0, 0, 0},
+	{"DEST", 20, 45, 1, FLD_ALIGN_LEFT, -1, 0, 0, 0},
+	{"GW", 20, 45, 1, FLD_ALIGN_LEFT, -1, 0, 0, 0},
+	{"STATE", 5, 23, 18, FLD_ALIGN_COLUMN, -1, 0, 0, 0},
+	{"AGE", 5, 9, 4, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"EXP", 5, 9, 4, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"PR ", 4, 9, 1, FLD_ALIGN_LEFT, -1, 0, 0, 0},
+	{"DIR", 1, 3, 2, FLD_ALIGN_CENTER, -1, 0, 0, 0},
+	{"PKTS", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"BYTES", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"RULE", 2, 4, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"LABEL", MIN_LABEL_SIZE, MIN_LABEL_SIZE, 1, FLD_ALIGN_LEFT, -1, 0, 0, 0},
+	{"STATES", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"EVAL", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"ACTION", 1, 8, 4, FLD_ALIGN_LEFT, -1, 0, 0, 0},
+	{"LOG", 1, 3, 2, FLD_ALIGN_LEFT, -1, 0, 0, 0},
+	{"QUICK", 1, 1, 1, FLD_ALIGN_LEFT, -1, 0, 0, 0},
+	{"KS", 1, 1, 1, FLD_ALIGN_LEFT, -1, 0, 0, 0},
+	{"IF", 4, 6, 1, FLD_ALIGN_LEFT, -1, 0, 0, 0},
+	{"INFO", 40, 80, 1, FLD_ALIGN_LEFT, -1, 0, 0, 0},
+	{"MAX", 3, 5, 2, FLD_ALIGN_RIGHT, -1, 0, FLD_FLAG_HIDDEN, 0},
+	{"RATE", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"AVG", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"PEAK", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"ANCHOR", 6, 16, 1, FLD_ALIGN_LEFT, -1, 0, FLD_FLAG_HIDDEN, 0},
+	{"QUEUE", 15, 30, 1, FLD_ALIGN_LEFT, -1, 0, 0, 0},
+	{"BW", 4, 5, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"SCH", 3, 4, 1, FLD_ALIGN_LEFT, -1, 0, 0, 0},
+	{"PRIO", 1, 4, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"DROP_P", 6, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"DROP_B", 6, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"QLEN", 4, 4, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"BORROW", 4, 6, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"SUSPENDS", 4, 6, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"P/S", 3, 7, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0},
+	{"B/S", 4, 7, 1, FLD_ALIGN_RIGHT, -1, 0, 0, 0}
 };
 
 
@@ -388,9 +394,11 @@ struct pf_altq_node {
 
 /* command prompt */
 
+void cmd_filter(void);
 void cmd_delay(void);
 void cmd_count(void);
 
+struct command cm_filter = {"Filter", cmd_filter};
 struct command cm_delay = {"Seconds to delay", cmd_delay};
 struct command cm_count = {"Number of lines to display", cmd_count};
 
@@ -701,18 +709,36 @@ read_states(void)
 		ps.ps_buf = (char *) state_buf;
 
 		if (ioctl(pf_dev, DIOCGETSTATES, &ps) < 0) {
-			errx(1, "DIOCGETSTATES");
+			msgprintf("Error Reading States (DIOCGETSTATES): %s",
+				    strerror(errno));
+			return (-1);
 		}
-		num_states = ps.ps_len / sizeof(pf_state_t);
+		num_states_all = ps.ps_len / sizeof(pf_state_t);
 
 		if (ps.ps_len < sbytes)
 			break;
 
-		alloc_buf(num_states);
+		alloc_buf(num_states_all);
 	}
 
-	for (n = 0; n<num_states; n++) {
-		state_ord[n] = n;
+	if (dumpfilter) {
+		int fd = open("state.dmp", O_WRONLY|O_CREAT|O_EXCL, 0);
+		if (fd > 0) {
+			write(fd, state_buf, ps.ps_len);
+			close(fd);
+		}
+	}
+
+	if (filter_prog.bf_insns == NULL) {
+		num_states =  num_states_all;
+		for (n = 0; n<num_states_all; n++)
+			state_ord[n] = n;
+	} else {
+		num_states = 0;
+		for (n = 0; n<num_states_all; n++)
+			if (bpf_filter(filter_prog.bf_insns, (char *)(&state_buf[n]),
+			    sizeof(pf_state_t), sizeof(pf_state_t)) > 0)
+				state_ord[num_states++] = n;
 	}
 
 	if (cachestates) {
@@ -761,7 +787,7 @@ print_header(void)
 	int end = dispstart + maxprint;
 
 	if (ioctl(pf_dev, DIOCGETSTATUS, &status)) {
-		warnx("DIOCGETSTATUS");
+		msgprintf("Error Reading status (DIOCGETSTATUS): %s", strerror(errno));
 		return (-1);
 	}
 	
@@ -779,6 +805,9 @@ print_header(void)
 		tbprintf(" no entries");
 	else
 		tbprintf(" %u-%u/%u", start, end, num_disp);
+
+	if (filter_prog.bf_insns != NULL)
+		tbprintf(" (%u)", num_states_all);
 	
 	if (curr_view) {
 		tbprintf(", View: %s", curr_view->name);
@@ -858,8 +887,7 @@ print_fld_host(field_def *fld, pf_state_host_t * h, int af)
 	else
 		tbprintf("[%u]", p);
 
-	print_fld_str(fld, tmp_buf);
-	tb_end();
+	print_fld_tb(fld);
 }
 
 void
@@ -908,9 +936,7 @@ print_fld_state(field_def *fld, unsigned int proto,
 		tbprintf("%u:%u", s1, s2);
 	}
 
-	print_fld_str(fld, tmp_buf);
-
-	tb_end();
+	print_fld_tb(fld);
 }
 
 int
@@ -1050,7 +1076,8 @@ read_anchor_rules(char *anchor)
 	strlcpy(pr.anchor, anchor, sizeof(pr.anchor));
 #endif
 	if (ioctl(pf_dev, DIOCGETRULES, &pr)) {
-		warn(anchor);
+		msgprintf("Error Reading Anchor %s (DIOCGETRULES): %s",
+			  *anchor ? anchor : "/", strerror(errno));
 		return (-1);
 	}
 
@@ -1061,7 +1088,8 @@ read_anchor_rules(char *anchor)
 	for (nr = 0; nr < num; ++nr) {
 		pr.nr = nr;
 		if (ioctl(pf_dev, DIOCGETRULE, &pr)) {
-			warn("DIOCGETRULE");
+			msgprintf("Error Reading Rule (DIOCGETRULE): %s",
+				  strerror(errno));
 			return (-1);
 		}
 #ifdef HAVE_RULESETS
@@ -1157,7 +1185,8 @@ read_rulesets(const char *path)
 	a->ref += len;
 
 	if (ioctl(pf_dev, DIOCGETRULESETS, &ruleset)) {
-		warn("DIOCGETRULESETS");
+		msgprintf("Error Reading Rulesets (DIOCGETRULESETS): %s",
+			  strerror(errno));
 		return (-1);
 	}
 
@@ -1176,7 +1205,8 @@ read_rulesets(const char *path)
 	for (nr = 0; nr < ns; ++nr) {
 		ruleset.nr = nr;
 		if (ioctl(pf_dev, DIOCGETRULESET, &ruleset)) {
-			warn("DIOCGETRULESET");
+			msgprintf("Error Reading Ruleset (DIOCGETRULESET): %s",
+				    strerror(errno));
 			return (-1);
 		}
 		*pre = '/';
@@ -1248,7 +1278,6 @@ read_rules(void)
 		    FLD_LABEL->max_width != mw) {
 			FLD_LABEL->norm_width = nw;
 			FLD_LABEL->max_width = mw;
-			fprintf(stderr, "update\n");
 			field_setup();
 			need_update = 1;
 		}
@@ -1494,7 +1523,7 @@ print_rule(struct pf_rule *pr)
 			tbprintf("!");
 #endif
 		tbprintf("%s", pr->ifname);
-		print_fld_str(FLD_IF, tmp_buf);
+		print_fld_tb(FLD_IF);
 	}
 #ifdef HAVE_MAX_STATES
 	if (pr->max_states)
@@ -1614,14 +1643,12 @@ print_rule(struct pf_rule *pr)
 		tbprintf("tagged %s ", pr->match_tagname);
 	}
 #endif
-	print_fld_str(FLD_RINFO, tmp_buf);
+	print_fld_tb(FLD_RINFO);
 
 #ifdef HAVE_RULESETS
 	/* XXX anchor field overloaded with anchor name */
 	print_fld_str(FLD_ANCHOR, (char *)pr->anchor);
 #endif
-	tb_end();
-
 	end_line();
 }
 
@@ -1726,14 +1753,16 @@ pfctl_update_qstats(struct pf_altq_node **root, int *inserts)
 	memset(&pq, 0, sizeof(pq));
 	memset(&qstats, 0, sizeof(qstats));
 	if (ioctl(pf_dev, DIOCGETALTQS, &pa)) {
-		warn("DIOCGETALTQS");
+		msgprintf("Error Reading Queues (DIOCGETALTQS): %s",
+			  strerror(errno));
 		return (-1);
 	}
 	num_queues = nr_queues = pa.nr;
 	for (nr = 0; nr < nr_queues; ++nr) {
 		pa.nr = nr;
 		if (ioctl(pf_dev, DIOCGETALTQ, &pa)) {
-			warn("DIOCGETALTQ");
+			msgprintf("Error Reading Queue (DIOCGETALTQ): %s",
+				  strerror(errno));
 			return (-1);
 		}
 		if (pa.altq.qid > 0) {
@@ -1742,7 +1771,8 @@ pfctl_update_qstats(struct pf_altq_node **root, int *inserts)
 			pq.buf = &qstats;
 			pq.nbytes = sizeof(qstats);
 			if (ioctl(pf_dev, DIOCGETQSTATS, &pq)) {
-				warn("DIOCGETQSTATS");
+				msgprintf("Error Reading Queue (DIOCGETQSTATS): %s",
+					  strerror(errno));
 				return (-1);
 			}
 			qstats.valid = 1;
@@ -1888,8 +1918,7 @@ print_queue(struct pf_altq_node *node)
 	for (d = 0; d < node->depth; d++)
 		tbprintf(" ");
 	tbprintf(node->altq.qname);
-	print_fld_str(FLD_QUEUE, tmp_buf);
-	tb_end();
+	print_fld_tb(FLD_QUEUE);
 
 	if (node->altq.scheduler == ALTQT_CBQ ||
 	    node->altq.scheduler == ALTQT_HFSC
@@ -1973,8 +2002,7 @@ print_queue(struct pf_altq_node *node)
 		else
 			tbprintf("%u", (unsigned int) pps);
 		
-		print_fld_str(FLD_PKTSPS, tmp_buf);
-		tb_end();
+		print_fld_tb(FLD_PKTSPS);
 		print_fld_bw(FLD_BYTESPS, bps);
 	}
 }
@@ -2042,19 +2070,21 @@ show_help(void)
 	erase();
 	mvprintw(line, 2, "pfTop Help");
 	line += 2;
-	mvprintw(line,    5, " c  - Toggle state cache");
-	mvprintw(line++, 40, " h  - Help (this page)");
-	mvprintw(line,    5, " n  - Set number of lines");
-	mvprintw(line++, 40, " o  - next sort Order");
-	mvprintw(line,    5, " p  - Pause display");
-	mvprintw(line++, 40, " r  - Reverse sort order");
-	mvprintw(line,    5, " s  - Set update interval");
-	mvprintw(line++, 40, " v  - next View");
-	mvprintw(line,    5, " q  - Quit");
+	mvprintw(line,    5, " c  - toggle state Cache");
+	mvprintw(line++, 40, " f  - set state Filter");
+	mvprintw(line,    5, " h  - Help (this page)");
+	mvprintw(line++, 40, " n  - set Number of lines");
+	mvprintw(line,    5, " o  - next sort Order");
+	mvprintw(line++, 40, " p  - Pause display");
+	mvprintw(line,    5, " r  - Reverse sort order");
+	mvprintw(line++, 40, " s  - Set update interval");
+	mvprintw(line,    5, " v  - next View");
+	mvprintw(line++, 40, " q  - Quit");
 	line++;
-	mvprintw(line++, 5, "0-6 - select view directly");
+	mvprintw(line++, 5, "0-8 - select view directly");
 	mvprintw(line++, 5, "SPC - update immediately");
 	mvprintw(line++, 5, "^L  - refresh display");
+	mvprintw(line++, 5, "^G  - clear command entry line");
 	line++;
 	mvprintw(line++, 5, "cursor keys - scroll display");
 	line++;
@@ -2076,6 +2106,33 @@ show_help(void)
 	mvprintw(line++, 6, "press any key to continue ...");
 
 	while (getch() == ERR);
+
+}
+
+
+void
+cmd_filter(void)
+{
+	struct bpf_program p;
+
+	if (*cmdbuf == '\0') {
+		sf_freecode(&filter_prog);
+		free(filter_string);
+		filter_string = NULL;
+		return;
+	}
+
+	if (sf_compile(&p, cmdbuf, 1, 0)) {
+		msgprintf("Error: %s", sf_get_error());
+		return;
+	}
+
+	if (filter_prog.bf_insns != NULL)
+		sf_freecode(&filter_prog);
+
+	filter_prog = p;
+	free(filter_string);
+	filter_string = strdup(cmdbuf);
 }
 
 void
@@ -2089,6 +2146,7 @@ cmd_delay(void)
 	}
 }
 
+
 void
 cmd_count(void)
 {
@@ -2110,6 +2168,9 @@ keyboard_callback(int ch)
 		cachestates = !cachestates;
 		update_cache();
 		break;
+	case 'f':
+		command_set(&cm_filter, filter_string);
+		break;
 	case '?':
 		/* FALLTHROUGH */
 	case 'h':
@@ -2117,14 +2178,15 @@ keyboard_callback(int ch)
 		need_update = 1;
 		break;
 	case 'n':
-		command_set(&cm_count);
+		command_set(&cm_count, NULL);
 		break;
 	case 's':
-		command_set(&cm_delay);
+		command_set(&cm_delay, NULL);
 		break;
 	default:
 		return 0;
 	};
+
 	return 1;
 }
 
@@ -2149,6 +2211,7 @@ main(int argc, char *argv[])
 
 	char *orderstr = NULL;
 	char *viewstr = NULL;
+	char *filterstr = NULL;
 
 	int cachesize = DEFAULT_CACHE_SIZE;
 	int countmax = 0;
@@ -2156,7 +2219,7 @@ main(int argc, char *argv[])
 
 	int ch;
 
-	while ((ch = getopt(argc, argv, "abhirc:s:d:o:v:w:S:")) != -1) {
+	while ((ch = getopt(argc, argv, "abDhirc:s:d:f:o:v:w:S:")) != -1) {
 		switch (ch) {
 		case 'a':
 			maxstates = -1;
@@ -2166,6 +2229,9 @@ main(int argc, char *argv[])
 			if (cachesize < 0)
 				cachesize = 0;
 			break;
+		case 'D':
+			dumpfilter = 1;
+			break;
 		case 'd':
 			countmax = atoi(optarg);
 			if (countmax < 0)
@@ -2178,6 +2244,9 @@ main(int argc, char *argv[])
 			rawmode = 1;
 			interactive = 0;
 			break;
+		case 'f':
+			filterstr = optarg;
+			break;
 		case 'o':
 			orderstr = optarg;
 			break;
@@ -2220,6 +2289,20 @@ main(int argc, char *argv[])
 	else if (argc > 1)
 		usage();
 
+	if (filterstr) {
+
+		if (sf_compile(&filter_prog, filterstr, 1, 0)) {
+			fprintf(stderr, "Filter error: %s\n", sf_get_error());
+			return 1;
+		}
+		filter_string = strdup(filterstr);
+		if (dumpfilter) {
+			bpf_dump(&filter_prog, 0);
+			rawmode = 1;
+			interactive = 0;
+		}
+	}
+
 	initialize();
 
 	set_order(orderstr);
